#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <boost/scoped_array.hpp>
#include <boost/program_options.hpp>
#include <boost/timer.hpp>
//      PE-
#include <pe_lib/pe_bliss.h>
#include <pe_lib/pe_bliss_resources.h>
//   LZO1Z999
#include "../../lzo-2.06/include/lzo/lzo1z.h"
//    
#include "structs.h"
//    
#include "../unpacker/parameters.h"
//  ()
#include "unpacker.h"

using namespace pe_bliss;

//      PE  LZO
#ifndef _M_X64
#ifdef _DEBUG
#pragma comment(lib, "../../pe_bliss_1.0.0/Debug/pe_bliss.lib")
#pragma comment(lib, "../Debug/lzo-2.06.lib")
#else
#pragma comment(lib, "../../pe_bliss_1.0.0/Release/pe_bliss.lib")
#pragma comment(lib, "../Release/lzo-2.06.lib")
#endif
#else
#ifdef _DEBUG
#pragma comment(lib, "../../pe_bliss_1.0.0/x64/Debug/pe_bliss.lib")
#pragma comment(lib, "../x64/Debug/lzo-2.06.lib")
#else
#pragma comment(lib, "../../pe_bliss_1.0.0/x64/Release/pe_bliss.lib")
#pragma comment(lib, "../x64/Release/lzo-2.06.lib")
#endif
#endif

int main(int argc, char* argv[])
{	
	//      
	namespace po = boost::program_options;

	//  ,  
	//   
	boost::timer pack_timer;

	//  -   
	//  
	bool force_mode = false;
	//  
	bool repack_resources;
	//    
	bool rebuild_load_config;
	//  DOS-
	bool strip_dos_headers;
	//   
	unsigned long file_alignment;

	//   
	std::string input_file_name;
	//   
	std::string output_file_name;

	//  
	po::options_description visible_options("DXPack Packer 1.1\nCommand Line Options");

	try
	{
		//   
		//     (  )
		po::options_description cmdline;

		//out-file,o - ,    "--out-file"
		//    "-o"
		visible_options.add_options() 
			("out-file,o", po::value<std::string>(&output_file_name), "Output file name")
			("file-align,a", po::value<unsigned long>(&file_alignment)->default_value(512), "Packed file alignment")
			("strip-dos,s", po::value<bool>(&strip_dos_headers)->default_value(true), "Strip DOS headers")
			("repack-res,r", po::value<bool>(&repack_resources)->default_value(true), "Repack resources")
			("build-load-config,l", po::value<bool>(&rebuild_load_config)->default_value(true), "Rebuild Load Config directory")
			("force,f", "Force packing of possibly incorrect binaries")
			;

		cmdline.add(visible_options);
		//  -    
		cmdline.add_options()
			("image", po::value<std::string>(&input_file_name), "PE image to pack")
			;    

		// (        )
		po::positional_options_description desc_pos;
		desc_pos.add("image", 1);

		//  
		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).
			options(cmdline).positional(desc_pos).run(), vm);
		po::notify(vm);

		//      
		if(input_file_name.empty())
			throw std::runtime_error("No input file specified");

		//    
		if(vm.count("force"))
		{
			std::cout << "Force mode is active!" << std::endl;
			force_mode = true;
		}
	}
	catch(const std::exception& e)
	{
		// -    -   
		std::cout << e.what() << std::endl << std::endl;
		std::cout << visible_options << std::endl;
		system("pause");
		return 0;
	}


	std::cout << "Packing " << input_file_name << std::endl;
	//  -      argv   1
	std::auto_ptr<std::ifstream> file;
	file.reset(new std::ifstream(input_file_name, std::ios::in | std::ios::binary));
	if(!*file)
	{
		//     -     
		std::cout << "Cannot open " << argv[1] << std::endl;
		return -1;
	}

	try
	{
		//    32- PE-
		//  false,     
		//""   
		//    ,     
		pe_base image(*file, pe_properties_32(), false);
		file.reset(0); //    

		//,  .NET    
		if(image.is_dotnet() && !force_mode)
		{
			std::cout << ".NET image cannot be packed!" << std::endl;
			return -1;
		}

		//   ,  ,    
		{
			std::cout << "Entropy of sections: ";
			double entropy = entropy_calculator::calculate_entropy(image);
			std::cout << entropy << std::endl;
			// wasm.ru  ,   ,
			//  PE-    6.8
			// ,  ,  , 
			// ( )    
			//  ,    
			if(entropy > 6.8)
			{
				std::cout << "File has already been packed!" << std::endl;

				if(!force_mode)
					return -1;
			}
		}

		//   LZO
		if(lzo_init() != LZO_E_OK)
		{
			std::cout << "Error initializing LZO library" << std::endl;
			return -1;
		}

		std::cout << "Reading sections..." << std::endl;

		//   PE-
		const section_list& sections = image.get_image_sections();
		if(sections.empty())
		{
			//      ,   
			std::cout << "File has no sections!" << std::endl;
			return -1;
		}

		//    PE-
		packed_file_info basic_info = {0};
		//     
		basic_info.number_of_sections = sections.size();
		//   LOCK
		basic_info.lock_opcode = 0xf0;

		//    
		//    
		basic_info.original_import_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_IMPORT);
		basic_info.original_import_directory_size = image.get_directory_size(IMAGE_DIRECTORY_ENTRY_IMPORT);
		//    
		//    
		basic_info.original_resource_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE);
		basic_info.original_resource_directory_size = image.get_directory_size(IMAGE_DIRECTORY_ENTRY_RESOURCE);
		//    
		//    
		basic_info.original_relocation_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_BASERELOC);
		basic_info.original_relocation_directory_size = image.get_directory_size(IMAGE_DIRECTORY_ENTRY_BASERELOC);

		//  
		//     
		basic_info.original_load_config_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);

		//   
		basic_info.original_entry_point = image.get_ep();
		//     
		// 
		basic_info.total_virtual_size_of_sections = image.get_size_of_image();


		//,    
		// packed_section   
		std::string packed_sections_info;

		{
			//        
			packed_sections_info.resize(sections.size() * sizeof(packed_section));

			//""   ,      
			std::string raw_section_data;
			//  
			unsigned long current_section = 0;

			//  
			for(section_list::const_iterator it = sections.begin(); it != sections.end(); ++it, ++current_section)
			{
				//   
				const section& s = *it;


				{
					//  
					//      
					packed_section& info
						= reinterpret_cast<packed_section&>(packed_sections_info[current_section * sizeof(packed_section)]);

					// 
					info.characteristics = s.get_characteristics();
					//   
					info.pointer_to_raw_data = s.get_pointer_to_raw_data();
					//  
					info.size_of_raw_data = s.get_size_of_raw_data();
					//   
					info.virtual_address = s.get_virtual_address();
					//  
					info.virtual_size = s.get_virtual_size();

					//   (  8 )
					memset(info.name, 0, sizeof(info.name));
					memcpy(info.name, s.get_name().c_str(), s.get_name().length());
				}

				//  ,   
				if(s.get_raw_data().empty())
					continue;

				//    -     
				//   
				raw_section_data += s.get_raw_data();
			}

			//    ,   !
			if(raw_section_data.empty())
			{
				std::cout << "All sections of PE file are empty!" << std::endl;
				return -1;
			}

			packed_sections_info += raw_section_data;
		}


		// 
		section new_section;
		// - .rsrc ( )
		new_section.set_name(".rsrc");
		//  , , 
		new_section.readable(true).writeable(true).executable(true);
		//    
		std::string& out_buf = new_section.get_raw_data();


		// "" 
		//      LZO 
		//     
		//  
		//   lzo_align_t  , 
		//    
		//(   LZO)
		boost::scoped_array<lzo_align_t> work_memory(new lzo_align_t[LZO1Z_999_MEM_COMPRESS]);

		//  
		lzo_uint src_length = packed_sections_info.size();
		//       
		basic_info.size_of_unpacked_data = src_length;

		//  
		//(  )
		lzo_uint out_length = 0;

		//    
		//( -     LZO)
		out_buf.resize(src_length + src_length / 16 + 64 + 3);

		//  
		std::cout << "Packing data..." << std::endl;
		if(LZO_E_OK !=
			lzo1z_999_compress(reinterpret_cast<const unsigned char*>(packed_sections_info.data()),
			src_length,
			reinterpret_cast<unsigned char*>(&out_buf[0]),
			&out_length,
			work_memory.get())
			)
		{
			// -  , 
			std::cout << "Error compressing data!" << std::endl;
			return -1;
		}

		std::cout << "Packing complete..." << std::endl;

		//      
		basic_info.size_of_packed_data = out_length;
		//      
		//   , 
		//  
		out_buf.resize(out_length);
		//  ,   
		//    
		out_buf =
			//  basic_info
			std::string(reinterpret_cast<const char*>(&basic_info), sizeof(basic_info))
			// 
			+ out_buf;

		//,     
		if(out_buf.size() >= src_length)
		{
			std::cout << "File is incompressible!" << std::endl;

			if(!force_mode)
				return -1;
		}


		//   TLS,    
		std::auto_ptr<tls_info> tls;
		if(image.has_tls())
		{
			std::cout << "Reading TLS..." << std::endl;
			tls.reset(new tls_info(get_tls_info(image)));
		}


		//   ,    
		//  
		exported_functions_list exports;
		export_info exports_info;
		if(image.has_exports())
		{
			std::cout << "Reading exports..." << std::endl;
			exports = get_exported_functions(image, exports_info);
		}


		//   Image Load Config,    
		std::auto_ptr<image_config_info> load_config;
		if(image.has_config() && rebuild_load_config)
		{
			std::cout << "Reading Image Load Config..." << std::endl;

			try
			{
				load_config.reset(new image_config_info(get_image_config(image)));
			}
			catch(const pe_exception& e)
			{
				image.remove_directory(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);
				std::cout << "Error reading load config directory: " << e.what() << std::endl;
			}
		}
		else
		{
			image.remove_directory(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);
		}

		{
			//     
			//  PE-
			const section& first_section = image.get_image_sections().front();
			//      ( )
			new_section.set_virtual_address(first_section.get_virtual_address());

			//     
			//  PE-
			const section& last_section = image.get_image_sections().back();
			//    
			DWORD total_virtual_size = 
				//   
				last_section.get_virtual_address()
				//    
				+ pe_utils::align_up(last_section.get_virtual_size(), image.get_section_alignment())
				//    
				- first_section.get_virtual_address();

			//    
			resource_directory new_root_dir;

			if(image.has_resources() && repack_resources)
			{
				std::cout << "Repacking resources..." << std::endl;

				//    ( )
				resource_directory root_dir = get_resources(image);
				//     
				//  
				pe_resource_viewer res(root_dir);
				pe_resource_manager new_res(new_root_dir);

				try
				{
					//    
					//  ,  ID
					pe_resource_viewer::resource_id_list icon_id_list(res.list_resource_ids(pe_resource_viewer::resource_icon_group));
					pe_resource_viewer::resource_name_list icon_name_list(res.list_resource_names(pe_resource_viewer::resource_icon_group));
					//    ,  ,   
					if(!icon_name_list.empty())
					{
						//        (  0)
						//        ,    list_resource_languages
						//        ,    get_icon_by_name (   )
						//      
						resource_cursor_icon_writer(new_res).add_icon(
							resource_cursor_icon_reader(res).get_icon_by_name(icon_name_list[0]),
							icon_name_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_icon_group, icon_name_list[0]).at(0));
					}
					else if(!icon_id_list.empty()) //    ,     ID
					{
						//        (  0)
						//        ,    list_resource_languages
						//        ,    get_icon_by_id_lang
						//      
						resource_cursor_icon_writer(new_res).add_icon(
							resource_cursor_icon_reader(res).get_icon_by_id(icon_id_list[0]),
							icon_id_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_icon_group, icon_id_list[0]).at(0));
					}
				}
				catch(const pe_exception&)
				{
					// -   , ,  ,
					//   
				}

				try
				{
					//  ,  ID
					pe_resource_viewer::resource_id_list manifest_id_list(res.list_resource_ids(pe_resource_viewer::resource_manifest));
					if(!manifest_id_list.empty()) //  
					{
						//        (  0)
						//     
						new_res.add_resource(
							res.get_resource_data_by_id(pe_resource_viewer::resource_manifest, manifest_id_list[0]).get_data(),
							pe_resource_viewer::resource_manifest,
							manifest_id_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_manifest, manifest_id_list[0]).at(0)
							);
					}
				}
				catch(const pe_exception&)
				{
					// -   ,
					//   
				}

				try
				{
					//     ,  ID
					pe_resource_viewer::resource_id_list version_info_id_list(res.list_resource_ids(pe_resource_viewer::resource_version));
					if(!version_info_id_list.empty()) //    
					{
						//           (  0)
						//       
						new_res.add_resource(
							res.get_resource_data_by_id(pe_resource_viewer::resource_version, version_info_id_list[0]).get_data(),
							pe_resource_viewer::resource_version,
							version_info_id_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_version, version_info_id_list[0]).at(0)
							);
					}
				}
				catch(const pe_exception&)
				{
					// -   ,
					//   
				}
			}


			//   PE-
			image.get_image_sections().clear();

			//  
			image.realign_file(file_alignment);

			//      
			//       
			section& added_section = image.add_section(new_section);
			//     
			image.set_section_virtual_size(added_section, total_virtual_size);


			std::cout << "Creating imports..." << std::endl;

			//    kernel32.dll
			import_library kernel32;
			kernel32.set_name("kernel32.dll"); //  

			//  
			imported_function func;
			func.set_name("LoadLibraryA"); // 
			kernel32.add_import(func); //   

			//  
			func.set_name("GetProcAddress");
			kernel32.add_import(func); // 

			//   (RVA)  load_library_a
			//  packed_file_info,     
			//  , ?
			DWORD load_library_address_rva = pe_base::rva_from_section_offset(added_section,
				offsetof(packed_file_info, load_library_a));

			//    
			//   (import address table)
			kernel32.set_rva_to_iat(load_library_address_rva);

			//   
			imported_functions_list imports;
			//    
			imports.push_back(kernel32);

			//  
			import_rebuilder_settings settings;
			//Original import address table    ( )
			settings.build_original_iat(false);
			//  IAT    ,
			//  (load_library_address_rva)
			settings.save_iat_and_original_iat_rvas(true, true);
			//      
			settings.set_offset_from_section_start(added_section.get_raw_data().size());

			//      ,
			//    
			//   
			if(!new_root_dir.get_entry_list().empty())
				settings.enable_auto_strip_last_section(false);

			// 
			rebuild_imports(image, imports, added_section, settings);

			// ,  ,  
			if(!new_root_dir.get_entry_list().empty())
				rebuild_resources(image, new_root_dir, added_section, added_section.get_raw_data().size());


			//    TLS
			if(tls.get())
			{
				//     
				// 
				//      ,
				//     
				packed_file_info* info = reinterpret_cast<packed_file_info*>(&added_section.get_raw_data()[0]);

				//   
				// TLS
				info->original_tls_index_rva = tls->get_index_rva();

				//    TLS-,   
				//       
				if(!tls->get_tls_callbacks().empty())
					info->original_rva_of_tls_callbacks = tls->get_callbacks_rva();

				//     TLS
				//  -        tls_index
				// packed_file_info
				tls->set_index_rva(pe_base::rva_from_section_offset(added_section, offsetof(packed_file_info, tls_index)));
			}
		}


		//    
		//   TLS-
		DWORD first_callback_offset = 0;

		{
			// 
			section unpacker_section;
			// - kaimi.ru
			unpacker_section.set_name("kaimi.ru");
			//  ,   
			unpacker_section.readable(true).executable(true).writeable(true);

			{
				std::cout << "Writing unpacker stub, size = " << sizeof(unpacker_data) << " bytes" << std::endl;
				//     
				std::string& unpacker_section_data = unpacker_section.get_raw_data();
				//   
				//     
				//unpacker.h,     main.cpp
				unpacker_section_data = std::string(reinterpret_cast<const char*>(unpacker_data), sizeof(unpacker_data));

				//    
				// 
				*reinterpret_cast<DWORD*>(&unpacker_section_data[original_image_base_offset]) = image.get_image_base_32();
				*reinterpret_cast<DWORD*>(&unpacker_section_data[original_image_base_no_fixup_offset]) = image.get_image_base_32();

				//       ,
				//         
				//    ,   , 
				// packed_file_info
				*reinterpret_cast<DWORD*>(&unpacker_section_data[rva_of_first_section_offset]) = image.get_image_sections().at(0).get_virtual_address();
			}

			//   
			section& unpacker_added_section = image.add_section(unpacker_section);

			if(tls.get() || image.has_exports() || image.has_reloc() || load_config.get())
			{
				//     
				//     
				//( ,       
				//    PE)
				unpacker_added_section.get_raw_data().resize(sizeof(unpacker_data));
			}


			//    TLS
			if(tls.get())
			{
				std::cout << "Rebuilding TLS..." << std::endl;

				//     
				//     
				std::string& data = unpacker_added_section.get_raw_data();

				// ,     IMAGE_TLS_DIRECTORY32
				DWORD directory_pos = data.size();
				//    
				// sizeof(DWORD)   ,  
				//IMAGE_TLS_DIRECTORY32    4-  
				data.resize(data.size() + sizeof(IMAGE_TLS_DIRECTORY32) + sizeof(DWORD));

				//  TLS  ...
				if(!tls->get_tls_callbacks().empty())
				{
					//  
					//  TLS-
					// 1    DWORD
					first_callback_offset = data.size();
					data.resize(data.size() + sizeof(DWORD) * (tls->get_tls_callbacks().size() + 1));

					//     (ret 0xC),
					//  
					*reinterpret_cast<DWORD*>(&data[first_callback_offset]) =
						image.rva_to_va_32(pe_base::rva_from_section_offset(unpacker_added_section, empty_tls_callback_offset));

					//   
					//  TLS-
					tls->set_callbacks_rva(pe_base::rva_from_section_offset(unpacker_added_section, first_callback_offset));

					//    packed_file_info,  
					//     ,
					//    
					reinterpret_cast<packed_file_info*>(&image.get_image_sections().at(0).get_raw_data()[0])->new_rva_of_tls_callbacks = tls->get_callbacks_rva();
				}
				else
				{
					//  ,     
					tls->set_callbacks_rva(0);
				}

				//  ,     
				//   
				tls->clear_tls_callbacks();

				//   
				//     
				tls->set_raw_data_start_rva(pe_base::rva_from_section_offset(unpacker_added_section, data.size()));
				//    
				tls->recalc_raw_data_end_rva();

				// TLS
				// ,       
				//    (  ,  )
				// ,         
				rebuild_tls(image, *tls, unpacker_added_section, directory_pos, false, false, tls_data_expand_raw, true, false);

				//    
				//  
				unpacker_added_section.get_raw_data() += tls->get_raw_data();
				//     "kaimi.ru"
				//  SizeOfZeroFill  TLS
				image.set_section_virtual_size(unpacker_added_section, data.size() + tls->get_size_of_zero_fill());

				//,        
				if(!image.has_reloc() && !image.has_exports() && !load_config.get())
					pe_utils::strip_nullbytes(unpacker_added_section.get_raw_data());

				//    (  )
				image.prepare_section(unpacker_added_section);
			}


			//    -   
			// ,    
			image.set_ep(image.rva_from_section_offset(unpacker_added_section, 0));
		}

		if(load_config.get())
		{
			std::cout << "Repacking load configuration..." << std::endl;

			section& unpacker_section = image.get_image_sections().at(1);

			//   Lock-
			load_config->clear_lock_prefix_list();
			load_config->add_lock_prefix_rva(pe_base::rva_from_section_offset(image.get_image_sections().at(0), offsetof(packed_file_info, lock_opcode)));

			//         "kaimi.ru"
			//   SE Handler',    Lock-  
			rebuild_image_config(image, *load_config, unpacker_section, unpacker_section.get_raw_data().size(), true, true, true, !image.has_reloc() && !image.has_exports());
		}

		//    
		if(image.has_reloc())
		{
			std::cout << "Creating relocations..." << std::endl;

			//      
			relocation_table_list reloc_tables;

			section& unpacker_section = image.get_image_sections().at(1);

			{
				relocation_table table;
				//    
				//       
				//,        
				// ,    
				table.set_rva(unpacker_section.get_virtual_address());

				//    original_image_base_offset 
				// parameters.h 
				table.add_relocation(relocation_entry(original_image_base_offset, IMAGE_REL_BASED_HIGHLOW));

				//    
				reloc_tables.push_back(table);
			}

			//    TLS
			if(tls.get())
			{
				//    TLS
				//   
				DWORD tls_directory_offset = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_TLS)
					- image.section_from_directory(IMAGE_DIRECTORY_ENTRY_TLS).get_virtual_address();
				
				//   ,      TLS    
				// original_image_base_offset
				relocation_table table;
				table.set_rva(image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_TLS));
				//    StartAddressOfRawData,
				//EndAddressOfRawData  AddressOfIndex
				//     
				table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_TLS_DIRECTORY32, StartAddressOfRawData)), IMAGE_REL_BASED_HIGHLOW));
				table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_TLS_DIRECTORY32, EndAddressOfRawData)), IMAGE_REL_BASED_HIGHLOW));
				table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_TLS_DIRECTORY32, AddressOfIndex)), IMAGE_REL_BASED_HIGHLOW));

				//  TLS-
				if(first_callback_offset)
				{
					//      AddressOfCallBacks
					//     
					table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_TLS_DIRECTORY32, AddressOfCallBacks)), IMAGE_REL_BASED_HIGHLOW));
					table.add_relocation(relocation_entry(static_cast<WORD>(tls->get_callbacks_rva() - table.get_rva()), IMAGE_REL_BASED_HIGHLOW));
				}

				reloc_tables.push_back(table);
			}

			if(load_config.get())
			{
				//   IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG,       ,
				//        PE-
				DWORD config_directory_offset = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG)
					- image.section_from_directory(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG).get_virtual_address();

				//   ,      TLS    
				// original_image_base_offset  TLS
				relocation_table table;
				table.set_rva(image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG));

				if(load_config->get_security_cookie_va())
					table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_LOAD_CONFIG_DIRECTORY32, SecurityCookie)), IMAGE_REL_BASED_HIGHLOW));

				if(load_config->get_se_handler_table_va())
					table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_LOAD_CONFIG_DIRECTORY32, SEHandlerTable)), IMAGE_REL_BASED_HIGHLOW));

				table.add_relocation(relocation_entry(static_cast<WORD>(offsetof(IMAGE_LOAD_CONFIG_DIRECTORY32, LockPrefixTable)), IMAGE_REL_BASED_HIGHLOW));
				reloc_tables.push_back(table);
			}

			// ,    
			//   
			rebuild_relocations(image, reloc_tables, unpacker_section, unpacker_section.get_raw_data().size(), true, !image.has_exports());
		}

		if(image.has_exports())
		{
			std::cout << "Repacking exports..." << std::endl;

			section& unpacker_section = image.get_image_sections().at(1);

			//       "kaimi.ru"
			rebuild_exports(image, exports_info, exports, unpacker_section, unpacker_section.get_raw_data().size(), true);
		}

		//    
		//      
		//  ,   
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_IAT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_SECURITY);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_DEBUG);

		//  ,   
		//  ,    12 ,    
		//    12  
		//image.strip_data_directories(16 - 4); // -   WinXP
		//   ,  - 
		image.strip_stub_overlay();


		std::string base_file_name;

		if(output_file_name.empty())
		{
			//  PE-
			//      
			base_file_name = input_file_name;
			std::string dir_name;
			std::string::size_type slash_pos;
			if((slash_pos = base_file_name.find_last_of("/\\")) != std::string::npos)
			{
				dir_name = base_file_name.substr(0, slash_pos + 1); //  
				base_file_name = base_file_name.substr(slash_pos + 1); //  
			}

			//    packed_ + __
			//    ,  
			// ,   
			base_file_name = dir_name + "packed_" + base_file_name;
		}
		else
		{
			base_file_name = output_file_name;
		}


		// 
		std::ofstream new_pe_file(base_file_name.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);
		if(!new_pe_file)
		{
			//     -  
			std::cout << "Cannot create " << base_file_name << std::endl;
			return -1;
		}

		// PE-
		// DOS-,    NT-
		//(     true)
		//  SizeOfHeaders -     
		rebuild_pe(image, new_pe_file, strip_dos_headers, false);

		// ,    
		std::cout << "Packed image was saved to " << base_file_name << std::endl;

		std::cout << "Resulting sections entropy: " << entropy_calculator::calculate_entropy(image) << std::endl;
		std::cout << "Finished in " << pack_timer.elapsed() << " sec" << std::endl;
	}
	catch(const pe_exception& e)
	{
		//  -     
		//    
		std::cout << "Error: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
